import os
import cgi
import webbrowser
import time
import requests
import urllib.parse
import importlib.util
import logging
import json
import keycloak
from uuid import uuid4
import presalytics_doc_converter
import presalytics_ooxml_automation
import presalytics_story
import presalytics.lib.exceptions
import presalytics.lib.constants as cnst
import presalytics.client.auth


logger = logging.getLogger(__name__)


class Client(object):
    """ Main PresalyticsClient base object """
    def __init__(
            self,
            config=None,
            config_file=None,
            delegate_login=False,
            **kwargs):
        try:
            self.username = presalytics.CONFIG['USERNAME']
        except KeyError:
            raise presalytics.lib.exceptions.MissingConfigException("Mandatory configuration variable PRESALYTICS_USERNAME is missing from configuration.  Please reconfigure and retry.")
        try:
            self.password = presalytics.CONFIG['PASSWORD']
            self.direct_grant = True
        except KeyError:
            self.password = None
            self.direct_grant = False
        try:
            self.client_id = presalytics.CONFIG['CLIENT_ID']
        except KeyError:
            self.client_id = cnst.DEFAULT_CLIENT_ID
        try:
            self.client_secret = presalytics.CONFIG['CLIENT_SECRET']
            self.confidential_client = True
        except KeyError:
            self.client_secret = None
            self.confidential_client = False

        try:
            self.site_host = presalytics.CONFIG["HOSTS"]["SITE"]
        except KeyError:
            self.site_host = cnst.HOST
        self.login_sleep_interval = 5  # seconds
        self.login_timeout = 60  # seconds
        self._delegate_login = delegate_login
        self.oidc = keycloak.KeycloakOpenID(
            server_url=cnst.OIDC_AUTH_HOST,
            realm_name=cnst.OIDC_REALM,
            client_id=self.client_id,
            verify=True
        )
        self.token_util = presalytics.client.auth.TokenUtil()
        self.token_util.token = self.refresh_token()

        doc_converter_api_client = DocConverterApiClientWithAuth(self, **kwargs)
        self.doc_converter = presalytics_doc_converter.DefaultApi(api_client=doc_converter_api_client)
        ooxml_automation_api_client = OoxmlAutomationApiClientWithAuth(self, **kwargs)
        self.ooxml_automation = presalytics_ooxml_automation.DefaultApi(api_client=ooxml_automation_api_client)
        story_api_client = StoryApiClientWithAuth(self, **kwargs)
        self.story = presalytics_story.DefaultApi(api_client=story_api_client)

    def login(self):
        try:
            if self.direct_grant:
                keycloak_token = self.oidc.token(username=self.username, password=self.password)
            else:
                keycloak_token = self._get_new_token_browser()
        except keycloak.exceptions.KeycloakGetError as e:
            if e.response_code == 404:
                logger.error("Received reponse code 404 from api authentication.  Indicates username does not exist.  Please recheck config and try again.")
            elif e.response_code == 401:
                logger.error("Received 401 response code from api authentication.  This user's access to this resource is unauthorized")
            elif e.response_code >= 500:
                logger.error("Server error.  Please try again later.")
            else:
                logger.error(e.error_message)
            raise presalytics.lib.exceptions.MissingConfigException(e.error_message)
        self.token_util.process_keycloak_token(keycloak_token)
        return self.token_util.token

    def delegated_login(self, original_token, audience=None):
        """ Requires developer account and authorized client credentials """
        self.oidc.decode_token(original_token, cnst.JWT_KEY)
        kwargs = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "subject_token": original_token
        }
        if audience is not None:
            kwargs.update(
                {
                    "audience": audience
                }
            )
        keycloak_token = self.oidc.token(**kwargs)
        self.token_util.process_keycloak_token(keycloak_token)
        return self.token_util.token["access_token"]

    def _get_new_token_browser(self):
        """
        Opens browser in on presalytics.io and prompts for user login.
        Retrieves authorization code from website and obtains api token
        """
        api_otp = uuid4()
        query = {
            "api_otp": api_otp,
            "client_id": self.client_id
        }
        query_string = '?{}'.format(urllib.parse.urlencode(query))
        url = urllib.parse.urljoin(self.site_host, urllib.parse.urljoin(cnst.LOGIN_PATH, query_string))
        webbrowser.open_new_tab(url) 
        auth_code = None
        payload = {
            "username": self.username,
            "api_otp": str(api_otp),
            "client_id": self.client_id
        }

        auth_code_url = urllib.parse.urljoin(self.site_host, cnst.API_CODE_URL)
        interval = 0
        while True:
            response = requests.post(auth_code_url, json=payload)
            if response.status_code != 200:
                if interval <= self.login_timeout:
                    interval += self.login_sleep_interval
                    time.sleep(self.login_sleep_interval)
                else:
                    raise presalytics.lib.exceptions.LoginTimeout()
            else:
                data = json.loads(response.content)
                break
        auth_code = data["authorization_code"]
        token = self.oidc.token(username=self.username, grant_type="authorization_code", code=auth_code, redirect_uri=cnst.REDIRECT_URI)
        return token

    def refresh_token(self):
        if self.token_util.is_api_access_token_expired():
            if self.token_util.is_api_refresh_token_expired():
                if not self._delegate_login:  # do not automatically log in at init
                    self.login()
                else:
                    self.token_util.token = None
            else:
                try:
                    refresh_token = self.token_util.token["refresh_token"]
                    self.token_util.token = self.oidc.refresh_token(refresh_token)
                except keycloak.exceptions.KeycloakGetError:
                    self.login()
            self.token_util._put_token_file()
        return self.token_util.token

    def get_auth_header(self):
        self.refresh_token()
        auth_header = {
            "Authorization": "Bearer " + self.token_util.token["access_token"]
        }
        return auth_header

    def download_file(self, story_id, ooxml_automation_id, download_folder=None, filename=None, **kwargs):
        response, status, headers = self.story.story_id_file_ooxmlautomationid_get_with_http_info(story_id, ooxml_automation_id, _preload_content=False)
        if download_folder is None:
            download_folder = os.getcwd()
        if filename is None:
            cd_header = headers.get('Content-Disposition')
            _, params = cgi.parse_header(cd_header)
            filename = params["filename"]
        filepath = os.path.join(download_folder, filename)
        with open(filepath, 'wb') as f:
            f.write(response.data)


class DocConverterApiClientWithAuth(presalytics.client.auth.AuthenticationMixIn, presalytics_doc_converter.api_client.ApiClient):
    def __init__(self, parent: Client, **kwargs):
        presalytics.client.auth.AuthenticationMixIn.__init__(self, parent, **kwargs)
        presalytics_doc_converter.api_client.ApiClient.__init__(self)
        self.update_configuration()


class OoxmlAutomationApiClientWithAuth(presalytics.client.auth.AuthenticationMixIn, presalytics_ooxml_automation.api_client.ApiClient):
    def __init__(self, parent: Client, **kwargs):
        presalytics.client.auth.AuthenticationMixIn.__init__(self, parent, **kwargs)
        presalytics_ooxml_automation.api_client.ApiClient.__init__(self)
        self.update_configuration()


class StoryApiClientWithAuth(presalytics.client.auth.AuthenticationMixIn, presalytics_story.api_client.ApiClient):
    def __init__(self, parent: Client, **kwargs):
        presalytics.client.auth.AuthenticationMixIn.__init__(self, parent, **kwargs)
        presalytics_story.api_client.ApiClient.__init__(self)
        self.update_configuration()



import typing
import time
import urllib.parse
import requests
import presalytics.client.api
import presalytics.story.components
import presalytics.client
import presalytics.lib.exceptions
import presalytics.story.outline


class OoxmlEndpointMap(object):
    BASE_URL = "https://api.presalytics.io/ooxml-automation"

    CHART = "Charts"
    CONNECTION_SHAPE = "ConnectionShapes"
    DOCUMENT = "Documents"
    GROUP = "Groups"
    IMAGE = "Images"
    SHAPE = "Shapes"
    SHAPETREE = "ShapeTrees"
    SLIDE = "Slides"
    TABLE = "Tables"
    THEME = "Themes"

    def __init__(self, endpoint, baseurl: str = None):
        if endpoint not in OoxmlEndpointMap.__dict__.values():
            raise presalytics.lib.exceptions.ValidationError("{0} is not a valid endpoint ID".format(endpoint))
        self.endpoint_id = endpoint
        if not baseurl:
            self.baseurl = OoxmlEndpointMap.BASE_URL
        else:
            self.baseurl = baseurl
        self.root_url = urllib.parse.urljoin(self.baseurl, endpoint)

    def get_id_url(self, id):
        return urllib.parse.urljoin(self.root_url, id)

    def get_svg_url(self, id):
        return urllib.parse.urljoin(self.root_url, "/Svg/" + id)
    
    def get_xml_url(self, id):
        return urllib.parse.urljoin(self.root_url, "/OpenOfficeXml/" + id)


class OoxmlEditorWidget(presalytics.story.components.WidgetBase):
    __component_kind__ = 'ooxml-xml-editor'

    __plugins__ = [
        {
            'name': 'external_scripts',
            'type': 'script',
            'config': {
                'approved_scripts_key': 'ooxml'
            }
        }
    ]

    def __init__(self, name: str, story_id: str, ooxml_id: str, endpoint_map: OoxmlEndpointMap, data: typing.Dict):
        self.story_id = story_id
        self.ooxml_id = ooxml_id
        self.name = name
        self.endpoint_map = endpoint_map
        self.svg_html = self.update(data)
        self.outline_widget = self.serialize()

    def update(self, data: typing.Dict) -> str:
        client = presalytics.client.api.Client()
        auth_header = client.get_auth_header()
        xml_url = self.endpoint_map.get_xml_url(self.ooxml_id)
        xml_response = requests.get(xml_url, headers=auth_header)
        if xml_response.status_code != 200:
            raise presalytics.lib.exceptions.ApiException(default_exception=xml_response.content)
        new_xml = self.update_xml(xml_response.text, data)
        put_data = {
            "id": self.ooxml_id,
            "openOfficeXml": new_xml,
            "type": self.endpoint_map.endpoint_id

        }
        xml_update_response = requests.put(xml_url, put_data, headers=auth_header)
        if xml_response.status_code != 200:
            raise presalytics.lib.exceptions.ApiException(default_exception=xml_update_response.content)
        svg_data = self.get_svg(id)
        return svg_data

    def to_html(self, **kwargs):
        return self.svg_html

    def get_svg(self, id, timeout_iterator=0) -> str:
        svg_url = self.endpoint_map.get_svg_url(id)
        client = presalytics.client.api.Client()
        auth_header = client.get_auth_header()
        response = requests.get(svg_url, headers=auth_header)
        svg_data = response.text
        if response.status_code != 200:
            raise presalytics.lib.exceptions.ApiException(default_exception=response.content)
        if response.text.startswith("Temp data"):
            if timeout_iterator > 5:
                raise presalytics.lib.exceptions.ApiException()
            else:
                time.sleep(2)
                svg_data = self.get_svg(id, timeout_iterator)
        return svg_data
                

    @classmethod
    def deseriailize(cls, component, **kwargs):
        return cls(component.name,
                   component.data["story_id"],
                   component.data["ooxml_id"],
                   component.data["endpoint_id"])

    def serialize(self, **kwargs):
        data = {
            "story_id": self.story_id,
            "ooxml_id": self.ooxml_id,
            "endpoint_id": self.endpoint_map.endpoint_id
        }
        widget = presalytics.story.outline.Widget(
            name=self.name,
            data=data,
            kind=self.__component_kind__
        )
        return widget

    def update_xml(self, old_xml: str, data: typing.Dict):
        raise NotImplementedError


class OoxmlFileWidget(presalytics.story.components.WidgetBase):
    def __init__(self, filename: str, endpoint_id: OoxmlEndpointMap, ):
        self.filename = filename




class classproperty(property):
    def __get__(self, obj, objtype=None):
        return super(classproperty, self).__get__(objtype)

    def __set__(self, obj, value):
        super(classproperty, self).__set__(type(obj), value)

    def __delete__(self, obj):
        super(classproperty, self).__delete__(type(obj))



""" Constants for api.presalytics.io client """
HOST = "https://api.presaltyics.io"
OIDC_WELL_KNOWN = "https://auth.presalytics.io/auth/realms/presalytics/.well-known/openid-configuration"
OIDC_TOKEN = "https://auth.presalytics.io/auth/realms/presalytics/protocol/openid-connect/token"
PORT = 443
OIDC_AUTH_HOST = "https://auth.presalytics.io/auth/"
OIDC_AUTH_PORT = 443
PUBLIC_KEY = 'MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsBQPBlYfkG2OQv6iE52HpWuCBxwVheGaesBOeNZFQV3NmIrH36tmXaqqOvn9JW8eGGo4a6tBfLGEqTb08/avLOypK1fni2ayh9UAKv9ajfBDbdqs5Qmoe3gr9WIrtjopxlUV6xabKyFVlD41vHFM9fTXAG8gRrPPbGmel081Egjjfqs6xM76Rtzz5E5puEpz1dWk7eClNhyL2W2gVAc8ldj6bZhjou3h+qn5k8rN4v/p98uUazselxcOArJ1QRpFdNuyqzUjAKsA6UmdM478bU9T68GUrMqo/mPIC7W3yevCKLGRwrFuE/ewCWRYfVpFcvwEHNsgNn9iqeAQY8Ce9QIDAQAB'
JWT_KEY = "-----BEGIN PUBLIC KEY-----\n{0}\n-----END PUBLIC KEY-----".format(PUBLIC_KEY)
REDIRECT_URI = "https://presalytics.io/login-success"
OIDC_REALM = "presalytics"
DEFAULT_CLIENT_ID = "python-client"
TOKEN_FILE = "token.json"
LOCALHOST_SERVER = "127.0.0.1"
LOCALHOST_PORT = 8052
LOGIN_PATH = "/accounts/login/"
API_CODE_URL = "/user/client-get-auth-code/"



from matplotlib import pyplot as plt
import presalytics
import presalytics.lib
import presalytics.lib.widgets
import presalytics.lib.widgets.matplotlib

fig, ax = plt.subplots()

ax.plot([3,1,4,1,5], 'ks-', mec='w', mew=5, ms=20)
ax.set_title("test plot")


test_plot_1 = presalytics.lib.widgets.matplotlib.MatplotlibFigure(fig, 'test_plot_1', module=__name__)



import typing
import presalytics.lib.plugins.base
import presalytics.lib.exceptions

script_map = {
    'd3': 'https://cdnjs.cloudflare.com/ajax/libs/d3/5.12.0/d3.min.js',
    'd3v3': 'https://cdnjs.cloudflare.com/ajax/libs/d3/3.5.17/d3.min.js',
    'reveal': {
        'base': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/js/reveal.min.js',
        'marked': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/markdown/marked.js',
        'markdown': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/markdown/markdown.min.js',
        'highlight': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/highlight/highlight.min.js',
        'math': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/math/math.min.js',
        'zoom': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/zoom-js/zoom.min.js',
        'notes': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/notes/notes.min.js',
        'print': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/print-pdf/print-pdf.min.js'

    },
    'mpld3': 'https://presalytics.io/static/mpld3/mpld3.min.js',
    'ooxml': 'https://presalytics.io/static/ooxml/ooxml.js'
}


styles_map = {
    'reveal': {
        'base': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/reveal.min.css',
        'themes': {
            'beige': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/beige.min.css',
            'black': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/black.min.css',
            'blood': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/blood.min.css',
            'league': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/league.min.css',
            'moon': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/moon.min.css',
            'night': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/night.min.css',
            'serif': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/serif.min.css',
            'simple': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/simple.min.css',
            'sky': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/sky.min.css',
            'solarized': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/solarized.min.css',
            'white': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/white.min.css'
        },
        'print': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/print/pdf.min.css'
    }
}


class AttrDict(dict):
    """ Dictionary subclass whose entries can be accessed by attributes
        (as well as normally).
    """
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    @staticmethod
    def from_nested_dict(data):
        """ Construct nested AttrDicts from nested dictionaries. """
        if not isinstance(data, dict):
            return data
        else:
            return AttrDict({key: AttrDict.from_nested_dict(data[key])
                            for key in data})

    def flatten(self) -> typing.Dict:
        return AttrDict._flatten(self, '', {})

    @staticmethod
    def _flatten(current, key, result) -> typing.Dict:
        if isinstance(current, dict):
            for k in current:
                new_key = "{0}.{1}".format(key, k) if len(key) > 0 else k
                AttrDict._flatten(current[k], new_key, result)
        else:
            result[key] = current
        return result


ALLOWED_SCRIPTS = AttrDict(script_map)


ALLOWED_STYLES = AttrDict(styles_map)


class ApprovedExternalLinks(presalytics.lib.plugins.base.StylePlugin):
    __plugin_name__ = 'external_links'

    def to_style(self, config, **kwargs):
        key = config['approved_styles_key']
        link = ALLOWED_STYLES.flatten()[key]
        if link is None:
            message = "Key {0} does not reference a link in the APPROVED_LINKS dictionary".format(key)
            raise presalytics.lib.exceptions.MissingConfigException(message)
        return '<link rel="stylesheet" href="{0}"/>'.format(link)


class ApprovedExternalScripts(presalytics.lib.plugins.base.ScriptPlugin):
    __plugin_name__ = 'external_scripts'

    def to_script(self, config, **kwargs):
        key = config['approved_scripts_key']
        link = ALLOWED_SCRIPTS.flatten()[key]
        if link is None:
            message = "Key {0} does not reference a link in the APPROVED_SCRIPTS dictionary".format(key)
            raise presalytics.lib.exceptions.MissingConfigException(message)
        return '<script type="text/javascript" src="{0}"></script>'.format(link)



import os
import shutil
import lxml.html
import webbrowser
import logging
import collections
import typing
import tempfile
import lxml
import lxml.html
import lxml.etree
import lxml.html.builder as E
import datetime
import presalytics
import presalytics.lib
import presalytics.lib.plugins
import presalytics.lib.plugins.base
import presalytics.lib.templates
import presalytics.lib.templates.base
import presalytics.story.util
import presalytics.story.server
import presalytics.story.components
import presalytics.lib.exceptions
if typing.TYPE_CHECKING:
    from presalytics.story.outline import StoryOutline, Page, Plugin
    from presalytics.story.components import PageTemplateBase, ComponentBase, WidgetBase, ThemeBase


logger = logging.getLogger('presalytics.story.revealer')


class Revealer(presalytics.story.components.Renderer):
    """
    This class renders 'Story Outines' to reveal.js presentations
    """
    story_outline: 'StoryOutline'
    base: lxml.etree.Element
    links: typing.Sequence[lxml.etree.Element]
    plugins: typing.List[typing.Dict]

    __component_kind__ = 'revealer'

    def __init__(
            self,
            story_outline: 'StoryOutline',
            **kwargs):

        self.story_outline = story_outline
        self.story_outline.validate()
        self.base = self.make_base()
        reveal_plugin_config = {
            'type': 'script',
            'name': 'reveal',
            'config': {}
        }
        self.update_outline_from_instances()
        self.plugins = [reveal_plugin_config]
        self.get_component_implicit_plugins()
        outline_plugins = presalytics.lib.plugins.base.PluginManager.get_plugins_from_nested_dict(story_outline.to_dict())
        self.plugins.extend(outline_plugins)
        self.plugin_mgr = presalytics.lib.plugins.base.PluginManager(self.plugins)

    @classmethod
    def deserialize(cls, component: 'StoryOutline', **kwargs):
        return cls(component, **kwargs)

    def serialize(self) -> 'StoryOutline':
        self.update_outline_from_instances()
        return self.story_outline

    def update_outline_from_instances(self, sub_dict: typing.Dict = None):
        if not sub_dict:
            sub_dict = self.story_outline.to_dict()
        if sub_dict:
            for key, val in sub_dict.items():
                if key in ["widgets", "themes", "pages"]:
                    if isinstance(val, list):
                        for list_item in val:
                            if isinstance(list_item, dict):
                                if "kind" in list_item:
                                    class_key = key.rstrip("s") + "." + list_item["kind"]
                                    klass = presalytics.COMPONENTS.get(class_key)
                                    if klass:
                                        if "name" in list_item:
                                            instance_key = class_key + "." + list_item["name"]
                                            inst = presalytics.COMPONENTS.get_instance(instance_key)
                                            if inst:
                                                self.set_outline_data_from_instance(inst)
                if isinstance(val, dict):
                    if len(val.keys()) > 0:
                        self.update_outline_from_instances(val)
                if isinstance(val, list):
                    for list_item in val:
                        if isinstance(list_item, dict):
                            self.update_outline_from_instances(list_item)

    def get_component_implicit_plugins(self, sub_dict: typing.Dict = None):
        if not sub_dict:
            sub_dict = self.story_outline.to_dict()
        if sub_dict:
            for key, val in sub_dict.items():
                if key in ["widgets", "themes", "pages"]:
                    if isinstance(val, list):
                        for list_item in val:
                            if isinstance(list_item, dict):
                                if "kind" in list_item:
                                    class_key = key.rstrip("s") + "." + list_item["kind"]
                                    klass = presalytics.COMPONENTS.get(class_key)
                                    if klass:
                                        if len(klass.__plugins__) > 0:
                                            self.plugins.extend(klass.__plugins__)
                                                
                if isinstance(val, dict):
                    if len(val.keys()) > 0:
                        self.get_component_implicit_plugins(val)
                if isinstance(val, list):
                    for list_item in val:
                        if isinstance(list_item, dict):
                            self.get_component_implicit_plugins(list_item)

    def set_outline_data_from_instance(self, inst: 'ComponentBase'):
        if inst.__component_type__ == 'widget':
            self.set_widget_outline_data(inst)
        if inst.__component_type__ == 'page':
            self.set_page_outline_data(inst)
        if inst.__component_type__ == 'theme':
            self.set_theme_outline_data(inst)

    def set_theme_outline_data(self, inst: 'ThemeBase'):
        theme_index = None
        for t in range(0, len(self.story_outline.themes)):
            if inst.name == self.story_outline.themes[t].name:
                theme_index = t
            if theme_index:
                break
        theme_outline = inst.serialize()
        if theme_index:
            self.story_outline.themes[theme_index] = theme_outline    

    def set_page_outline_data(self, inst: 'PageTemplateBase'):
        page_index = None
        for p in range(0, len(self.story_outline.pages)):
            if inst.name == self.story_outline.pages[p].name:
                page_index = p
            if page_index:
                break
        page_outline = inst.serialize()
        if page_index:
            self.story_outline.pages[page_index] = page_outline


    def set_widget_outline_data(self, inst: 'WidgetBase'):
        widget_index = None
        page_index = None
        for p in range(0, len(self.story_outline.pages)):
            for w in range(0, len(self.story_outline.pages[p].widgets)):
                widget = self.story_outline.pages[p].widgets[w]
                if widget.name == inst.name:
                    page_index = p
                    widget_index = w
                if page_index:
                    break
            if page_index:
                break
        w_outline = inst.serialize()
        if isinstance(page_index, int) and isinstance(widget_index, int):
            self.story_outline.pages[page_index].widgets[widget_index] = w_outline  

    def update_outline(self):
        raise NotImplementedError

    def make_base(self):
        base = lxml.etree.Element("div", attrib={"class": "reveal"})
        lxml.etree.SubElement(base, "div", attrib={"class": "slides"})
        return base

    def package_as_standalone(self):
        pres = self.render()
        body = E.BODY()
        body.append(pres)
        body = self.strip_unauthorized_scripts(body)
        for scripts in self.plugin_mgr.get_scripts():
            lxml_script = lxml.html.fragment_fromstring(scripts)
            body.append(lxml_script)
        head = E.HEAD()
        for link in self.plugin_mgr.get_styles():
            lxml_link = lxml.html.fragment_fromstring(link)
            head.append(lxml_link)
        head = self.strip_unauthorized_scripts(head)
        html = E.HTML(
            head,
            body
        )
        return lxml.html.tostring(html, pretty_print=True)

    def render_plugin(self, plugin):
        plugin_type = plugin['type']
        plugin_config = plugin['config']
        plugin_class = self.plugin_loader.load(plugin_type)
        script_string = plugin_class().to_string(plugin_config)
        return lxml.html.fragment_fromstring(script_string)

    def strip_unauthorized_scripts(self, body):
        allowed_scripts = presalytics.lib.plugins.external.ALLOWED_SCRIPTS.flatten().values()
        script_elements = body.findall(".//script")
        for ele in script_elements:
            try:
                link = ele.get("src")
            except KeyError:
                ele.getparent().remove(ele)
            if link not in allowed_scripts:
                ele.getparent().remove(ele)
        return body

    def update_info(self):
        info = self.story_outline.info
        info.date_modified = datetime.datetime.now()

    def render(self):
        """
        This method returns a t
        """
        reveal_base = self.base
        for page in self.story_outline.pages:
            slide = lxml.etree.SubElement(reveal_base, "section")
            page_html = self.render_page(page)
            slide_fragment = lxml.html.fragment_fromstring(page_html)
            slide.append(slide_fragment)
        return reveal_base

    def render_page(self, page: 'Page') -> str:
        class_key = "page." + page.kind
        key = class_key + "." + page.name
        if presalytics.COMPONENTS.get_instance(key):
            page_instance = presalytics.COMPONENTS.get_instance(key)
        else:
            klass = presalytics.COMPONENTS.get(class_key)
            deserialize_method = getattr(klass, "deserialize", None)
            if callable(deserialize_method):
                page_instance = deserialize_method(page)
            else:
                message = "Page component instance or class (kind) {0} unavailable in component registry".format(key)
                raise presalytics.lib.exceptions.MissingConfigException(message)
        return page_instance.render()

    @staticmethod
    def make_local_folders(files_path=None):
        if files_path is None:
            files_path = os.getcwd()
        if not os.path.exists(files_path):
            os.mkdir(files_path)
        templates_path = os.path.join(files_path, "templates")
        if not os.path.exists(templates_path):
            os.mkdir(templates_path)
        static_files_path = os.path.join(files_path, "static")
        if not os.path.exists(static_files_path):
            os.mkdir(static_files_path)
        css_path = os.path.join(static_files_path, "css")
        if not os.path.exists(css_path):
            os.mkdir(css_path)
        js_path = os.path.join(static_files_path, "js")
        if not os.path.exists(js_path):
            os.mkdir(js_path)
        theme_path = os.path.join(css_path, "theme")
        if not os.path.exists(theme_path):
            os.mkdir(theme_path)

    def present(self, files_path=None, debug=True, port=8082, host='127.0.0.1'):
        if not files_path:
            files_path = tempfile.gettempdir()
        html = self.package_as_standalone().decode('utf-8')
        id = presalytics.story.util.to_title_case(self.story_outline.title)
        self.make_local_folders(files_path)
        template_folder = os.path.dirname(presalytics.lib.templates.base .__file__)
        shutil.copy(os.path.join(template_folder, "favicon.ico"), os.path.join(files_path, "static"))
        html_file = os.path.join(files_path, "templates", id + '.html')
        with open(html_file, 'w') as file:
            file.write(html)
        address = "http://{}:{}/story/{}".format(host, port, id)
        presalytics.story.server.Browser(address).start()
        presalytics.story.server.LocalServer(host=host, debug=debug, port=port, root_path=files_path).run()



import os
import typing
import jinja2
import presalytics.story.outline
import presalytics.story.components
import presalytics.lib.util as util
if typing.TYPE_CHECKING:
    from presalytics.story.components import WidgetBase
    from presalytics.story.outline import Page


class JinjaTemplateBuilder(presalytics.story.components.PageTemplateBase):
    __css__: typing.Sequence[str]
    __template_file__: str
    __template_paths__: typing.List[str]
    template_paths: typing.List[str]

    __template_paths__ = []

    def __init__(self, page: 'Page', **kwargs) -> None:
        super().__init__(page, **kwargs)
        pkg_templates = os.path.join(os.path.dirname(__file__), "html")
        self.template_paths = [pkg_templates]
        if len(self.__template_paths__) > 0:
            self.template_paths[0:0] = self.__template_paths__

    @classmethod
    def deserialize(cls, component, **kwargs):
        return cls(component, **kwargs)

    @util.classproperty
    def __plugins__(cls):
        plugin_list = []
        for id in cls.__css__:
            new_item = {
                'type': 'style',
                'name': 'local',
                'config': {
                    'css_file_id': id
                }
            }
            plugin_list.append(new_item)
        return plugin_list

    def serialize(self):
        updated_plugins = []
        for plugin_data in self.__plugins__:
            updated_plugins.append(presalytics.story.outline.Plugin(**plugin_data))
        self.outline_page.plugins = updated_plugins
        return self.outline_page

    def get_template_name(self):
        """
        Requires subclasses have either a "__template___" property or override this method
        """
        if self.__template_file__:
            return self.__template_file__
        else:
            raise NotImplementedError

    def render(self, **kwargs) -> str:
        template = self.load_jinja_template()
        context = {
            "widgets": self.widgets,
            "widget_index": 0
        }
        if self.outline_page.additional_properties:
            context.update(self.outline_page.additional_properties)
        return template.render(**context)

    def load_jinja_template(self) -> jinja2.Template:
        loader = jinja2.FileSystemLoader(self.template_paths)
        env = jinja2.Environment(loader=loader)
        return env.get_template(self.get_template_name())


class TitleWithSingleItem(JinjaTemplateBuilder):
    __component_kind__ = 'TitleWithSingleItem'
    __css__ = ['single_item_grid']
    __template_file__ = 'title_with_single_widget.html'



import logging

logger = logging.getLogger('presalytics.lib.exceptions')


class PresalyticsBaseException(Exception):
    def __init__(self, *args, **kwargs):
        default_message = "An Error occur in the presalytics module.  Consult log for stacktrace."
        if not (args or kwargs):
            args = (default_message,)
            super().__init__(*args, **kwargs)


class MissingConfigException(PresalyticsBaseException):
    def __init__(self, message):
        if message is None:
            message = "Configuration file missing.  Please please a config.ini file in working directory"
        super().__init__(message)


class MisConfiguredTokenException(PresalyticsBaseException):
    def __init__(self):
        message = "Authenication token is missing or malformed.  Please acquire new token and try again."
        super().__init__(message)


class InvalidTokenException(PresalyticsBaseException):
    def __init__(self):
        message = "Configuration file missing.  Please please a config.ini file in working directory"
        super().__init__(message)


class LoginTimeout(PresalyticsBaseException):
    def __init__(self):
        message = "User took too long to login on website to obtain token.  Please try again."
        super().__init__(message)


class ValidationError(PresalyticsBaseException):
    def __init__(self, message):
        if not message:
            message = "Validation Error occured"
        super().__init__(message)


class ApiException(PresalyticsBaseException):
    def __init__(self, default_exception=None):
        if default_exception is not None:
            _attrs = [a for a in dir(default_exception) if not a.startswith('__')]
            for key in _attrs:
                setattr(self, key, getattr(default_exception, key))

    def __str__(self):
        """Custom error messages for exception"""
        try:
            error_message = "({0})\n"\
                            "Reason: {1}\n".format(self.status, self.reason)

            if self.body:
                error_message += "HTTP response body: {0}\n".format(self.body)

            return error_message
        except Exception:
            return "An unknown error occured.  Please set default_exception to learn more."



# Automatically configure default logger
import os
import environs
import logging
import presalytics.lib
import presalytics.lib.logger
import presalytics.lib.plugins
import presalytics.lib.plugins.base
import presalytics.lib.config_loader
import presalytics.story
import presalytics.story.components

env = environs.Env()
env.read_env()


# A comma-separated list of paths to search of config.py, plugin classes, and component classes
autodiscover_paths = env.list('AUTODISCOVER_PATHS', [])

CONFIG = presalytics.lib.config_loader.load_config(additional_paths=autodiscover_paths)

file_logger = CONFIG.get("USE_LOGGER", True)
log_level = CONFIG.get("LOG_LEVEL", logging.DEBUG)

presalytics.lib.logger.configure_logger(log_level=log_level, file_logger=file_logger)

registry_kwargs = {
    'show_errors': False,
    'autodiscover_paths': autodiscover_paths
}

PLUGINS = presalytics.lib.plugins.base.PluginRegistry(**registry_kwargs)

COMPONENTS = presalytics.story.components.ComponentRegistry(**registry_kwargs)




import json
import logging
import pkg_resources
import sys
import weakref
import dateutil
import dateutil.parser
import six
import datetime
import presalytics.lib.exceptions
import presalytics.lib.constants
import typing
if typing.TYPE_CHECKING:
    from presalytics.client.api import Client


logger = logging.getLogger(__name__)


class TokenUtil(object):
    def __init__(self, token=None, token_file=None):
        if token_file is None:
            self.token_file = presalytics.lib.constants.TOKEN_FILE
        else:
            self.token_file = token_file
        try:
            self._load_token_file()
        except Exception:
            pass
        if token is not None:
            try:
                self.token = {
                    'access_token': token['access_token'],
                    'access_token_expire_time': token["access_token_expire_time"],
                    'refresh_token': token["refresh_token"],
                    'refresh_token_expire_time': token["refresh_token_expire_time"]
                }
                self._put_token_file()
            except Exception:
                raise presalytics.lib.exceptions.MisConfiguredTokenException()

    def is_api_access_token_expired(self):
        try:
            expire_datetime = dateutil.parser.parse(self.token['access_token_expire_time'])
            if expire_datetime < datetime.datetime.now():
                return True
            return False
        except Exception:
            return True

    def is_api_refresh_token_expired(self):
        try:
            expire_datetime = dateutil.parser.parse(self.token['refresh_token_expire_time'])
            if expire_datetime < datetime.datetime.now():
                return True
            return False
        except Exception:
            return True

    def _load_token_file(self):
        self.token = self.load_token_from_file(self.token_file)

    def _put_token_file(self):
        self.put_token_file(self.token, self.token_file)

    def process_keycloak_token(self, keycloak_token):
        access_token_expire_time = datetime.datetime.now() + datetime.timedelta(seconds=keycloak_token['expires_in'])
        refresh_token_expire_time = datetime.datetime.now() + datetime.timedelta(seconds=keycloak_token['refresh_expires_in'])

        self.token = {
            'access_token': keycloak_token['access_token'],
            'refresh_token': keycloak_token['refresh_token'],
            'access_token_expire_time': access_token_expire_time.isoformat(),
            'refresh_token_expire_time': refresh_token_expire_time.isoformat()
        }

        return self.token

    @staticmethod
    def put_token_file(token, token_filepath):
        with open(token_filepath, mode='w+') as newtoken:
            json.dump(token, newtoken)

    @staticmethod
    def load_token_from_file(token_filepath):
        with open(token_filepath, 'r') as token_file:
            token = json.load(token_file)
        return token


class AuthenticationMixIn(object):
    def __init__(self, parent: 'Client', config=None, config_file=None, ignore_api_exceptions=False, **kwargs):
        self.parent: 'Client' = weakref.ref(parent)
        self._ignore_api_exceptions = ignore_api_exceptions
        super(AuthenticationMixIn, self).__init__(**kwargs)

    def call_api(
            self, resource_path, method,
            path_params=None, query_params=None, header_params=None,
            body=None, post_params=None, files=None,
            response_type=None, auth_settings=None, async_req=None,
            _return_http_data_only=None, collection_formats=None,
            _preload_content=None, _request_timeout=None, _host=None):
        """
        Overriding call_api to force token check, refresh on each api call,
        rather than at class initialized (good for ipython notebooks)
        """

        auth_header = self.parent().get_auth_header()
        if header_params is not None:
            header_params.update(auth_header)
        else:
            header_params = auth_header
        try:
            return super(AuthenticationMixIn, self).call_api(
                resource_path, method, path_params,
                query_params, header_params, body, post_params, files, response_type,
                auth_settings, async_req, _return_http_data_only, collection_formats, _preload_content,
                _request_timeout, _host
            )
        except Exception as e:
            if type(e).__name__ == "ApiException":
                if self._ignore_api_exceptions:
                    return e.body, e.status, e.headers
                else:
                    raise presalytics.lib.exceptions.ApiException(default_exception=e)
            else:
                t, v, tb = sys.exc_info()
                six.reraise(t, v, tb)

    def update_configuration(self):
        """
        updates the configuration from the base api client to parameters contained in the config file.
        base api client must be initialized prior to calling this method (i.e., self.configuration cannot equal None),
        so that configuration can automatically inherit default values.
        Mostly used for debugging purposes, but self-hosted API endpoints may require injection of these parameters
        """
        if self.configuration is None:
            raise presalytics.lib.exceptions.MissingConfigException("Base API not yet configured, please reconstruct API initialization")
        self.user_agent = AuthenticationMixIn._get_user_agent
        if presalytics.CONFIG.get("HOSTS", None):
            self.set_host(presalytics.CONFIG.get('HOSTS'))

    @staticmethod
    def get_user_agent():
        try:
            VER = pkg_resources.require("presalytics")[0].version
        except Exception:
            VER = 'build'
        return "presalytics-python-client/{0}".format(VER)

    _get_user_agent = get_user_agent.__func__()

    def set_host(self, hosts_dict):
        for parent_cls in self.__class__.__bases__:
            module_name = parent_cls.__module__.split('.')[0]
            if module_name != 'presalytics':
                host_key = module_name.replace('presalytics_', '').upper()
                try:
                    self.configuration.host = hosts_dict[host_key]
                    break
                except KeyError:
                    pass



import jinja2
import presalytics.lib.exceptions
import typing


class JinjaPluginMakerMixin(object):
    template: typing.Optional[str]
    template = None

    def __init__(self, *args, **kwargs):
        super(JinjaPluginMakerMixin, self).__init__(*args, *kwargs)

    def render(self, config: typing.Dict):
        options = {
            "loader": jinja2.BaseLoader()
        }
        if config.get("jinja_options", None):
            options.update(config.pop("jinja_options"))
        template = jinja2.Environment(**options).from_string(self.get_template())
        return template.render(config)

    def get_template(self) -> str:
        if not self.template:
            raise presalytics.lib.exceptions.ValidationError("Plugins the use the JinjaPluginMakerMixin must have a defiined template attribute.")
        else:
            return self.template



import unittest
import os
import json
import presalytics.story.revealer
import presalytics.story.outline


class TestStory(unittest.TestCase):
    """
    Test module features thatr render stories to dashboards or other formats
    """
    def setUp(self):
        pass

    def test_deserailze_yaml(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'matplotlib-outline.yaml')
        outline = presalytics.story.outline.StoryOutline.import_yaml(test_file)
        self.assertTrue(isinstance(outline, presalytics.story.outline.StoryOutline))
        self.assertTrue(isinstance(outline.info, presalytics.story.outline.Info))
        self.assertTrue(isinstance(outline.pages[0], presalytics.story.outline.Page))
        self.assertTrue(isinstance(outline.pages[0].widgets[0], presalytics.story.outline.Widget))
        self.assertTrue(isinstance(outline.themes[0], presalytics.story.outline.Theme))
        self.assertEqual('Test-story', outline.title)

    def test_serialize_to_json(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'matplotlib-outline.yaml')
        outline = presalytics.story.outline.StoryOutline.import_yaml(test_file)
        json_str = outline.dump()
        new_dict = json.loads(json_str)
        self.assertEqual(new_dict["pages"][0]["widgets"][0]["data"]["plotName"], "test_plt_1")
        self.assertIsNotNone(json_str)

    def test_render_matplotlib_plot(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'matplotlib-outline.yaml')
        outline = presalytics.story.outline.StoryOutline.import_yaml(test_file)
        presalytics.story.revealer.Revealer(outline).present(files_path='/tmp')

    def test_plugins(self):
        from presalytics import PLUGINS

    def tearDown(self):
        pass



import pkgutil
import importlib
import traceback
import sys
import os
import inspect
import logging
import typing
import abc


logger = logging.getLogger('presalytics.lib.registry')


class RegistryBase(abc.ABC):
    registry: typing.Dict[str, typing.Type]
    show_errors = False

    def __init__(self, show_errors=False, autodiscover_paths=[], reserved_names: typing.List[str] = None, **kwargs):
        RegistryBase.show_errors = show_errors
        self.autodiscover_paths = autodiscover_paths
        self.registry = {}
        self.reserved_names = ["config.py", "setup.py"]
        try:
            if reserved_names:
                self.reserved_names.extend(reserved_names)
        except Exception:
            pass
        self.discover()

    @abc.abstractmethod
    def get_type(self, klass):
        raise NotImplementedError

    @abc.abstractmethod
    def get_name(self, klass):
        raise NotImplementedError

    @staticmethod
    def onerror(name):
        if RegistryBase.show_errors:
            print("Error importing module %s" % name)
            t, value, tb = sys.exc_info()
            traceback.print_tb(tb)

    def get(self, key):
        return self.registry.get(key, None)

    def get_classes_from_module(self, module):
        self.get_classes(module)
        for loader, name, is_pkg in pkgutil.walk_packages(module.__path__, onerror=RegistryBase.onerror):
            full_name = module.__name__ + '.' + name
            try:
                sub_module = importlib.import_module(full_name)
                if sub_module.__name__.startswith('presalytics'):
                    if is_pkg:
                        self.get_classes_from_module(sub_module)
                    else:
                        self.get_classes(sub_module)
            except Exception:
                pass

    def load_class(self, klass):
        if inspect.isclass(klass):
            klass_type = self.get_type(klass)
            if klass_type:
                try:
                    klass_name = self.get_name(klass)
                    if klass_name:
                        key = "{0}.{1}".format(klass_type, klass_name)
                        if key not in self.registry.keys():
                            self.registry[key] = klass
                except Exception:
                    message = "Unable to register class {0} with type {1}".format(klass.__name__, klass_type)
                    logger.error(message)
        else:
            try:
                message = "{0} is not a class".format(klass.__class__.__name__)
                logger.error(message)
            except Exception:
                pass

    def get_classes(self, module):
        for key, val in module.__dict__.items():
            if inspect.isclass(val) or isinstance(val, abc.ABC):
                self.load_class(val)

    def discover(self):
        current_path = os.getcwd()
        if current_path not in self.autodiscover_paths:
            self.autodiscover_paths.append(current_path)
        for path in self.autodiscover_paths:
            for name in os.listdir(path):
                if name.endswith(".py") and name not in self.reserved_names:
                    try:
                        mod_path = os.path.join(path, name)
                        mod_spec = importlib.util.spec_from_file_location(name, mod_path)
                        mod = importlib.util.module_from_spec(mod_spec)
                        mod_spec.loader.exec_module(mod)
                        self.get_classes(mod)
                    except Exception:
                        message = "Could not load classes from file {0}".format(name)
                        logger.error(message)
        for finder, name, ispkg in pkgutil.iter_modules():
            if name.startswith('presalytics'):
                mod = importlib.import_module(name)
                self.get_classes_from_module(mod)

    def register(self, klass):
        self.load_class(klass)



import presalytics.lib.plugins.base as ext
import presalytics.lib.plugins.jinja as jinja


class RevealConfigPlugin(ext.ScriptPlugin, jinja.JinjaPluginMakerMixin):
    __plugin_name__ = 'reveal'

    __dependencies__ = [
        {
            'type': 'script',
            'name': 'external_scripts',
            'config': {
                'approved_scripts_key': 'reveal.base'
            }
        },
        {
            'type': 'style',
            'name': 'external_links',
            'config': {
                'approved_styles_key': 'reveal.base'
            }

        }
    ]

    def to_script(self, config, **kwargs):
        reveal_config = self.default_config
        if config.get("reveal_params", None):
            reveal_config.update(config.pop("reveal_params"))
        config.update({'reveal_config': reveal_config})
        return self.render(config)

    template = '<script type="text/javascript">Reveal.initialize({{ reveal_config|tojson }});</script>'

    default_config = {
        'controls': True,  # Display presentation control arrows
        # Help the user learn the controls by providing hints, for example by
        # bouncing the down arrow when they first encounter a vertical slide
        'controlsTutorial': True,

        # Determines where controls appear, "edges" or "bottom-right" #
        'controlsLayout': 'bottom-right',

        # Visibility rule for backwards navigation arrows; "faded", "hidden"
        # or "visible"
        'controlsBackArrows': 'faded',

        # Display a presentation progress bar
        'progress': True,

        # Display the page number of the current slide
        'slideNumber': False,

        # Add the current slide number to the URL hash so that reloading the
        #page/copying the URL will return you to the same slide
        'hash': False,

        # Push each slide change to the browser history. Implies `hash: true`
        'history': False,

        # Enable keyboard shortcuts for navigation 
        'keyboard': True,

        # Enable the slide overview mode
        'overview': True,

        # Vertical centering of slides
        'center': True,

        # Enables touch navigation on devices with touch input
        'touch': True,

        # Loop the presentation
        'loop': False,

        # Change the presentation direction to be RTL
        'rtl': False,

        # See https://github.com/hakimel/reveal.js/#navigation-mode
        'navigationMode': 'default',

        # Randomizes the order of slides each time the presentation loads
        'shuffle': False,

        # Turns fragments on and off globally
        'fragments': True,

        # Flags whether to include the current fragment in the URL,
        # so that reloading brings you to the same fragment position #}
        'fragmentInURL': False,

        # Flags if the presentation is running in an embedded mode,
        # i.e. contained within a limited portion of the screen #}
        'embedded': False,

        # Flags if we should show a help overlay when the questionmark
        # key is pressed #}
        'help': True,

        # Flags if speaker notes should be visible to all viewers
        'showNotes': False,

        # Global override for autoplaying embedded media (video/audio/iframe)
        # - null: Media will only autoplay if data-autoplay is present
        # - true: All media will autoplay, regardless of individual setting
        # - false: No media will autoplay, regardless of individual setting
        'autoPlayMedia': 'null',

        # Global override for preloading lazy-loaded iframes
        # - null: Iframes with data-src AND data-preload will be loaded when within
        #   the viewDistance, iframes with only data-src will be loaded when visible
        # - true: All iframes with data-src will be loaded when within the viewDistance
        # - false: All iframes with data-src will be loaded only when visible #}
        'preloadIframes': 'null',

        # Number of milliseconds between automatically proceeding to the
        # next slide, disabled when set to 0, this value can be overwritten
        # by using a data-autoslide attribute on your slides #}
        'autoSlide': 0,

        # Stop auto-sliding after user input
        'autoSlideStoppable': True,

        # Use this method for navigation when auto-sliding
        'autoSlideMethod': 'Reveal.navigateNext',

        # Specify the average time in seconds that you think you will spend
        # presenting each slide. This is used to show a pacing timer in the
        # speaker view 
        'defaultTiming': 120,

        # Enable slide navigation via mouse wheel
        'mouseWheel': False,

        # Hide cursor if inactive
        'hideInactiveCursor': True,

        # Time before the cursor is hidden (in ms)
        'hideCursorTime': 5000,

        # Hides the address bar on mobile devices
        'hideAddressBar': True,

        # Opens links in an iframe preview overlay
        #  Add `data-preview-link` and `data-preview-link="false"` to customise each link
        # individually
        'previewLinks': False,

        # Transition style
        'transition': 'slide', # none/fade/slide/convex/concave/zoom

        # Transition speed
        'transitionSpeed': 'default', # default/fast/slow

        # Transition style for full page slide backgrounds
        'backgroundTransition': 'fade', # none/fade/slide/convex/concave/zoom

        # Number of slides away from the current that are visible
        'viewDistance': 3,

        # Parallax background image
        'parallaxBackgroundImage': '', # e.g. "'https://s3.amazonaws.com/hakim-static/reveal-js/reveal-parallax-1.jpg'"

        # Parallax background size
        'parallaxBackgroundSize': '', # CSS syntax, e.g. "2100px 900px"

        # Number of pixels to move the parallax background per slide
        # - Calculated automatically unless specified
        # - Set to 0 to disable movement along an axis 
        'parallaxBackgroundHorizontal': 'null',
        'parallaxBackgroundVertical': 'null',

        # The display mode that will be used to show slides
        'display': 'block',

        # Allow for responsive presenation formats
        'width': "100%",
        'height': "100%",
        'margin': 0,
        'minScale': 1,
        'maxScale': 1
    }

    #  template = """<script type="text/javascript"
    # Reveal.initialize({

    #     {# Display presentation control arrows #}
    #     controls: true,

    #     {# Help the user learn the controls by providing hints, for example by
    #      bouncing the down arrow when they first encounter a vertical slide #}
    #     controlsTutorial: true,

    #     {# Determines where controls appear, "edges" or "bottom-right" #}
    #     controlsLayout: 'bottom-right',

    #     {# Visibility rule for backwards navigation arrows; "faded", "hidden"
    #      or "visible" #}
    #     controlsBackArrows: 'faded',

    #     {# Display a presentation progress bar #}
    #     progress: true,

    #     {# Display the page number of the current slide #}
    #     slideNumber: false,

    #     {# Add the current slide number to the URL hash so that reloading the
    #      page/copying the URL will return you to the same slide #}
    #     hash: false,

    #     {# Push each slide change to the browser history. Implies `hash: true` #}
    #     history: false,

    #     {# Enable keyboard shortcuts for navigation #}
    #     keyboard: true,

    #     {# Enable the slide overview mode #}
    #     overview: true,

    #     {# Vertical centering of slides #}
    #     center: true,

    #     {# Enables touch navigation on devices with touch input #}
    #     touch: true,

    #     {# Loop the presentation #}
    #     loop: false,

    #     {# Change the presentation direction to be RTL #}
    #     rtl: false,

    #     {# See https://
    #     github.com/hakimel/reveal.js/#navigation-mode #}
    #     navigationMode: 'default',

    #     {# Randomizes the order of slides each time the presentation loads #}
    #     shuffle: false,

    #     {# Turns fragments on and off globally #}
    #     fragments: true,

    #     {# Flags whether to include the current fragment in the URL,
    #      so that reloading brings you to the same fragment position #}
    #     fragmentInURL: false,

    #     {# Flags if the presentation is running in an embedded mode,
    #      i.e. contained within a limited portion of the screen #}
    #     embedded: false,

    #     {# Flags if we should show a help overlay when the questionmark
    #      key is pressed #}
    #     help: true,

    #     {# Flags if speaker notes should be visible to all viewers #}
    #     showNotes: false,

    #     {# Global override for autoplaying embedded media (video/audio/iframe)
    #      - null: Media will only autoplay if data-autoplay is present
    #      - true: All media will autoplay, regardless of individual setting
    #      - false: No media will autoplay, regardless of individual setting
    #     autoPlayMedia: null,

    #     {# Global override for preloading lazy-loaded iframes
    #      - null: Iframes with data-src AND data-preload will be loaded when within
    #        the viewDistance, iframes with only data-src will be loaded when visible
    #      - true: All iframes with data-src will be loaded when within the viewDistance
    #      - false: All iframes with data-src will be loaded only when visible #}
    #     preloadIframes: null,

    #     {# Number of milliseconds between automatically proceeding to the
    #     next slide, disabled when set to 0, this value can be overwritten
    #     by using a data-autoslide attribute on your slides #}
    #     autoSlide: 0,

    #     {# Stop auto-sliding after user input #}
    #     autoSlideStoppable: true,

    #     {# Use this method for navigation when auto-sliding #}
    #     autoSlideMethod: Reveal.navigateNext,

    #     {# Specify the average time in seconds that you think you will spend
    #      presenting each slide. This is used to show a pacing timer in the
    #      speaker view #}
    #     defaultTiming: 120,

    #     {# Enable slide navigation via mouse wheel #}
    #     mouseWheel: false,

    #     {# Hide cursor if inactive #}
    #     hideInactiveCursor: true,

    #     {# Time before the cursor is hidden (in ms) #}
    #     hideCursorTime: 5000,

    #     {# Hides the address bar on mobile devices #}
    #     hideAddressBar: true,

    #     {# Opens links in an iframe preview overlay
    #      Add `data-preview-link` and `data-preview-link="false"` to customise each link
    #      individually #}
    #     previewLinks: false,

    #     {# Transition style #}
    #     transition: 'slide', {# none/fade/slide/convex/concave/zoom #}

    #     {# Transition speed #}
    #     transitionSpeed: 'default', {# default/fast/slow #}

    #     {# Transition style for full page slide backgrounds #}
    #     backgroundTransition: 'fade', {# none/fade/slide/convex/concave/zoom #}

    #     {# Number of slides away from the current that are visible #}
    #     viewDistance: 3,

    #     {# Parallax background image
    #     parallaxBackgroundImage: '', {# e.g. "'https://s3.amazonaws.com/hakim-static/reveal-js/reveal-parallax-1.jpg'" #}

    #     {# Parallax background size
    #     parallaxBackgroundSize: '', {# CSS syntax, e.g. "2100px 900px" #}

    #     {# Number of pixels to move the parallax background per slide
    #      - Calculated automatically unless specified
    #      - Set to 0 to disable movement along an axis #}
    #     parallaxBackgroundHorizontal: null,
    #     parallaxBackgroundVertical: null,

    #     {# The display mode that will be used to show slides #}
    #     display: 'block',

    #     {# Allow for responsive presenation formats #}
    #     width: "100%",
    #     height: "100%",
    #     margin: 0,
    #     minScale: 1,
    #     maxScale: 1

    #     });
    #     </script>

    # """




import os
from environs import Env

env = Env()
env.read_env()

PRESALYTICS = {
    'USERNAME': os.environ['PRESALYTICS_USERNAME'],
    'PASSWORD': os.environ['PRESALYTICS_PASSWORD'],
    'HOSTS': {}
}

try:
    ooxml_host = {
        'OOXML_AUTOMATION': os.environ['OOXML_AUTOMATION_HOST']
    }
    PRESALYTICS['HOSTS'].update(ooxml_host)
except KeyError:
    pass

try:
    story_host = {
        'STORY': os.environ['STORY_HOST']
    }
    PRESALYTICS['HOSTS'].update(story_host)
except KeyError:
    pass


try:
    site_host = {
        'SITE': os.environ['SITE_HOST']
    }
    PRESALYTICS['HOSTS'].update(site_host)
except KeyError:
    pass



"""
Utility functions for the presalytics.story module
"""
import re
import logging
import typing

logger = logging.getLogger('presalytics.story.util')


def to_snake_case(camel_case_str):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel_case_str)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def to_camel_case(snake_str):
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


def to_title_case(name_string):
    components = name_string.split('_')
    return ''.join(x[0].upper() + x[1:] for x in components)



import os
import time
import flask
import threading
import webbrowser


app = flask.Flask(__name__)


@app.route('/story/<id>')
def story(id):
    template_name = str(id) + '.html'
    return flask.render_template(template_name)


@app.route('/shutdown', methods=['POST'])
def shutdown():
    shutdown_func = flask.request.environ.get('werkzeug.server.shutdown')
    shutdown_func()
    return 'Server shutting down...'


class LocalServer(object):
    def __init__(self, host='127.0.0.1', debug=True, port=8080, root_path=None, **kwargs):
        self.host = host
        self.debug = debug
        self.port = port
        self.root_path = root_path
        if self.root_path is None:
            self.root_path = os.getcwd()

    def run(self):
        app.use_reloader = False
        app.root_path = self.root_path
        app.run(host=self.host, debug=self.debug, port=self.port)


class Browser(threading.Thread):
    def __init__(self, address, delay_time=2, *args, **kwargs):
        super(Browser, self).__init__(*args, **kwargs)
        self.address = address
        self.delay_time = delay_time

    def run(self):
        time.sleep(self.delay_time)
        webbrowser.open_new_tab(self.address)


if __name__ == '__main__':
    app.run(debug=True)



import logging

logger = logging.getLogger('presalytics.teller')
logger.info("Importing presalytics story teller")


class StoryWriter(object):
    """
    This class takes input of various MimeTypes and transforms them
    into standarized json object - a 'Story Outline' that the revealer class and can
    parse into a reveal.js presentation
    """
    def __init__(self):
        pass

    def create_from_document(self, document):

        dummy_dict = {}
        return dummy_dict



# coding: utf-8

"""
    Presalytics Python Client

    Tools to interfacing with the Presalytics.io API.
"""


from setuptools import setup, find_packages  # noqa: H301

NAME = "presalytics"
VERSION = "0.2.21"

# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["urllib3 >= 1.15", 
    "six >= 1.10", 
    "certifi", 
    "python-dateutil",
    "flask",
    "requests",
    "python-keycloak",
    "presalytics-doc-converter",
    "presalytics-ooxml-automation",
    "presalytics-story",
    "environs",
    "matplotlib",
    "lxml",
    "pyyaml",
    "mpld3"]

with open("README.md", "r") as fh:
    long_description = fh.read()
    

setup(
    name=NAME,
    version=VERSION,
    description="Prealytics Python Client",
    author_email="inquiries@presalytics.io",
    url="https://presalytics.io/docs",
    keywords=["Presalytics"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description_content_type="text/markdown",
    long_description=long_description
)



"""
Module helps serialize and deserialize presalytics story outlines to/from json yaml.
Allow downstream modules work seemlessly with json data

Note: this module leverage type-checking at runtime, and may cause ciruclar references
if you are incorporating type hints on presaltyics objects in modules that import this file.
The problem is solved using the TYPE_CHECKING boolean in the python's typing package.
"""
import json
import yaml
import inspect
import datetime
import dateutil.parser
import uuid
import abc
import typing
from presalytics.story.util import to_camel_case, to_snake_case
from presalytics.lib.exceptions import ValidationError


class OutlineEncoder(json.JSONEncoder):
    def default(self, obj):
        if issubclass(obj.__class__, OutlineBase):
            return obj.to_dict()
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        if isinstance(obj, uuid.uuid4):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


class OutlineBase(abc.ABC):
    __client_ver__: str
    __annotations__: typing.Dict
    __required__: typing.Sequence[str]

    __required__ = []

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            key_name = to_snake_case(key)
            val = self.deserialize_child(key_name, val)
            setattr(self, key_name, val)
        self.validate()

    def validate(self):
        for key in self.__required__:
            if not hasattr(self, key):
                err_message = 'Could not load {0} object, source data missing "{1}" key'.format(self.__class__.__name__, key)
                raise ValidationError(err_message)

    def deserialize_child(self, key_name, val):
        if key_name in self.__annotations__:
            sub_type = self.__annotations__[key_name]
            if inspect.isclass(sub_type) and not issubclass(sub_type, str):
                if issubclass(sub_type, OutlineBase):
                    val = sub_type(**val)
                if issubclass(sub_type, typing.Sequence):
                    try:
                        list_type = sub_type.__args__[0]
                        if issubclass(list_type, OutlineBase):
                            for x in range(0, len(val)):
                                val[x] = list_type(**val[x])
                    except TypeError:
                        pass
                if issubclass(sub_type, typing.Dict):
                    try:
                        val_type = sub_type.__args__[1]
                        if issubclass(val_type, OutlineBase):
                            for child_key, child_val in val.items():
                                updated_val = val_type({child_key, child_val})
                                val.update(child_key, updated_val)
                    except TypeError:
                        pass
                if issubclass(sub_type, datetime.datetime):
                    val = dateutil.parser.parse(val)
                if issubclass(sub_type, uuid.UUID):
                    val = uuid.UUID(val)
        return val

    @classmethod
    def deserialize(cls, json_obj):
        return cls(**json_obj)

    @classmethod
    def load(cls, json_str):
        return json.JSONDecoder(object_hook=cls.deserialize).decode(json)

    @classmethod
    def import_yaml(cls, yaml_file):
        with open(yaml_file, 'r') as file:
            obj = yaml.safe_load(file)
        return cls.deserialize(obj)

    def dump(self):
        return json.dumps(self, cls=OutlineEncoder)

    def to_dict(self):
        ret = {}
        for key, val in self.__dict__.items():
            ret_key = "{}".format(to_camel_case(key))
            ret_val = json.loads(json.dumps(val, cls=OutlineEncoder), encoding='utf-8')
            ret[ret_key] = ret_val
        return ret


class Info(OutlineBase):
    revision: str
    date_created: datetime.datetime
    date_modified: datetime.datetime
    created_by: str
    modified_by: str
    revision_notes: str


class Widget(OutlineBase):
    name: str
    kind: str
    data: typing.Dict[str, str]
    plugins: typing.Sequence[str]
    additional_properties: typing.Dict[str, str]

    __required__ = [
        'name',
        'data'
    ]


class Page(OutlineBase):
    name: str
    kind: str
    widgets: typing.Sequence[Widget]
    additional_properties: typing.Dict[str, str]
    plugins: typing.List[typing.Dict]

    __required__ = [
        'name'
    ]


class Theme(OutlineBase):
    name: str
    module_name: str
    additional_properties: typing.Dict[str, str]

    __required__ = [
        'name'
    ]


class Plugin(OutlineBase):
    type: str
    name: str
    config: typing.Dict

    __required__ = [
        'type',
        'config',
        'name'
    ]


class StoryOutline(OutlineBase):
    presalytics_story: str
    info: Info
    pages: typing.Sequence[Page]
    description: str
    title: str
    themes: typing.Sequence[Theme]

    __required__ = [
        'presalytics_story',
        'info',
        'pages',
        'title'
    ]



import pkgutil
import importlib
import os
import typing
import importlib.util


def load_config(additional_paths: typing.List[str] = []) -> typing.Dict:
    current_path = os.getcwd()
    additional_paths.append(current_path)
    config_dict = None
    for path in additional_paths:
        for name in os.listdir(path):
            if name == "config.py":
                config_path = os.path.join(path, name)
                config_spec = importlib.util.spec_from_file_location("config", config_path)
                config_mod = importlib.util.module_from_spec(config_spec)
                config_spec.loader.exec_module(config_mod)
                config_dict = getattr(config_mod, "PRESALYTICS", None)
                if not config_dict:
                    config_dict = config_mod.__dict__
                if config_dict:
                    break
        if config_dict:
            break
    if config_dict:
        return config_dict
    else:
        return {}
